/**
 * SPDX-FileCopyrightText: 2026 KDE Connect Team
 *
 * SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
 */

#pragma once

class QWindow;

void activateWindowForMacOS(QWindow *window);
